"""
Author: Artner Michael
13.06.2021
"""
from abc import ABC
from typing import Tuple, List, Callable, Optional, Iterator, Union

from qrogue.game.logic.actors.controllables import Controllable
from qrogue.game.logic.actors.controllables.qubit import QubitSet, DummyQubitSet
from qrogue.game.logic.base import StateVector, CircuitMatrix, QuantumSimulator, QuantumCircuit, UnitarySimulator
from qrogue.game.logic.collectibles import Collectible, Instruction, Key, MultiCollectible, \
    Qubit, Energy, Score, InstructionManager
from qrogue.util import CheatConfig, Config, Logger, GameplayConfig, QuantumSimulationConfig, Options, GateType, \
    LevelInfo


# from jkq import ddsim
class RoboProperties:
    @staticmethod
    def from_difficulty_level(difficulty_level: int, gate_list: Optional[List[Instruction]] = None) -> "RoboProperties":
        return RoboProperties(LevelInfo.get_default_num_of_qubits(difficulty_level),
                              LevelInfo.get_default_circuit_space(difficulty_level), gate_list)

    def __init__(self, num_of_qubits: int = 3, circuit_space: int = 5, gate_list: Optional[List[Instruction]] = None):
        self.__num_of_qubits = num_of_qubits
        self.__circuit_space = circuit_space
        self.__gate_list = gate_list

    @property
    def num_of_qubits(self) -> int:
        return self.__num_of_qubits

    @property
    def circuit_space(self) -> int:
        return self.__circuit_space

    @property
    def instruction_list(self) -> Optional[List[Instruction]]:
        if self.__gate_list is None: return None
        return [gate.copy() for gate in self.__gate_list]


class _Attributes:
    """
    A class that handles some attributes of a Robot.
    """
    # todo: Overhaul whole class?
    __DEFAULT_SPACE = 5
    __DEFAULT_MAX_ENERGY = 100
    __MIN_INIT_ENERGY = 1  # during initialization neither max_energy nor cur_energy must be below this value

    def __init__(self, qubits: QubitSet = DummyQubitSet(), space: int = __DEFAULT_SPACE,
                 max_energy: int = __DEFAULT_MAX_ENERGY, start_energy: int = None):
        """

        :param qubits: the QubitSet the Robot is currently using
        :param space: how many instructions the robot can put on their circuit
        :param max_energy: how much energy the robot can store at most
        :param start_energy: with how much energy the robot starts. If it is bigger than max_energy, max_energy will be
                             used instead of the specified number
        """
        if space is None:
            space = _Attributes.__DEFAULT_SPACE
        if max_energy is None:
            max_energy = _Attributes.__DEFAULT_MAX_ENERGY
        if start_energy is None:
            start_energy = max_energy

        assert space > 0
        assert max_energy > _Attributes.__MIN_INIT_ENERGY
        assert start_energy > _Attributes.__MIN_INIT_ENERGY

        self.__qubits: QubitSet = qubits
        self.__space: int = space
        self.__max_energy: int = max_energy
        self.__cur_energy: int = min(start_energy, max_energy)

    @property
    def num_of_qubits(self) -> int:
        """

        :return: number of qubits the Robot's QubitSet has
        """
        return self.__qubits.size

    @property
    def circuit_space(self) -> int:
        """

        :return: maximum number of Instructions the Robot can use in its Circuit
        """
        return self.__space

    @property
    def qubits(self) -> QubitSet:
        """

        :return: the QubitSet used by the Robot
        """
        return self.__qubits

    @property
    def cur_energy(self) -> int:
        """

        :return: the amount of energy the Robot currently has
        """
        return self.__cur_energy

    @property
    def max_energy(self) -> int:
        """

        :return: the maximum amount of energy the Robot can store
        """
        return self.__max_energy

    def add_qubits(self, additional_qubits: int = 1):
        """
        Adds the given number of additional qubits to the QubitSet.

        :param additional_qubits: how many qubits to add (defaults to 1)
        :return: None
        """
        self.__qubits = self.__qubits.add_qubits(additional_qubits)

    def update_circuit_space(self, new_space: int) -> bool:
        if new_space <= 0: return False

        self.__space = new_space
        return True

    def increase_energy(self, amount: int) -> int:
        """
        Increases current energy by the given amount up to maximum energy.

        :param amount: by how much we want to increase current energy
        :return: by how much current energy was actually increased
        """
        self.__cur_energy += amount
        if self.__cur_energy > self.__max_energy:
            overflow = self.__cur_energy - self.__max_energy
            amount -= overflow
            self.__cur_energy = self.__max_energy
        return amount

    def decrease_energy(self, amount: int) -> int:
        """
        Decreases current energy by the given amount at most to 0.

        :param amount: by how much we want to decrease current energy
        :return: by how much current energy was actually decreased
        """
        if CheatConfig.got_inf_resources():
            return 0  # no decrease in this case

        self.__cur_energy -= amount
        if self.__cur_energy < 0:
            # e.g. if we decrease by 6 then cur_energy is now -2, so actually we were only able to decrease it by 4
            amount += self.__cur_energy
            self.__cur_energy = -1
        return amount


class _Backpack:
    """
    Stores Instructions, Consumables and other Collectibles for a Robot to use.
    """

    __DEFAULT_CAPACITY: int = 5  # how many Instructions the Backpack can hold at once

    def __init__(self, capacity: Optional[int] = None, content: Optional[List[Instruction]] = None):
        """
        Backpack is a storage/management class for Collectibles.

        :param capacity: how many Instructions can be stored in this Backpack
        :param content: list of initially stored Instructions
        """
        if capacity is None:
            capacity = _Backpack.__DEFAULT_CAPACITY

        self.__capacity: int = capacity
        if content:
            # self.__capacity = max(len(content), capacity)
            self.__storage: List[Instruction] = content
        else:
            self.__capacity = capacity
            self.__storage: List[Instruction] = []
        self.__key_count: int = 0

    @property
    def capacity(self) -> int:
        """

        :return: how many Collectibles can be stored in this Backpack
        """
        return self.__capacity

    @property
    def used_capacity(self) -> int:
        """

        :return: number of items currently stored
        """
        return len(self.__storage)

    @property
    def key_count(self) -> int:
        """

        :return: number of Keys we currently have
        """
        if CheatConfig.got_inf_resources():
            return 999
        return self.__key_count

    def give_key(self, amount: int) -> bool:
        """
        Adds the given amount of Keys. Fails if amount is less or equal to 0.

        :param amount: how many Keys we want to add
        :return: True if the given amount of Keys where handed out successfully, False otherwise
        """
        if amount > 0:
            self.__key_count += amount
            return True
        return False

    def use_key(self) -> bool:
        """
        Uses one Key, i.e. decreases key count by 1.

        :return: True if we could successfully use a Key, False if it failed (e.g. no Key left)
        """
        if CheatConfig.got_inf_resources():
            return True

        if self.key_count > 0:
            self.__key_count -= 1
            return True
        return False

    def get(self, index: int) -> Optional[Instruction]:
        """
        Returns the Instruction at the provided index in the storage if index is valid. Otherwise returns None.

        :param index: index of the Instruction we want to get
        :return: the Instruction at the given index or None
        """
        if 0 <= index < self.used_capacity:
            return self.__storage[index]
        return None

    def add(self, instruction: Instruction, force: bool = False) -> bool:
        """
        Adds an Instruction to the backpack if possible (i.e. if there is still space left).

        :param instruction: the Instruction to add
        :param force: whether we force instruction to be added, defaults to False
        :return: True if there is enough capacity left to store the Instruction, False otherwise
        """
        if self.used_capacity < self.__capacity or force:
            self.__storage.append(instruction)
            return True
        return False

    def remove(self, instruction: Instruction) -> bool:
        """
        Removes a given Instruction from the backpack if it's actually stored in it.

        :param instruction: the Instruction we want to remove
        :return: True if the Instruction is in the backpack and we were able to remove it, False otherwise
        """
        for i in range(len(self.__storage)):
            if self.__storage[i] == instruction:
                self.__storage.remove(instruction)
                return True
        if Config.debugging():
            Logger.instance().error("Reached a line in Backpack.remove() that I think should not be reachable ("
                                    "although it has no game-consequences if I'm wrong).", show=False, from_pycui=False)
        try:
            self.__storage.remove(instruction)
            return True
        except ValueError:
            return False

    def set_instructions(self, gate_list: List[Instruction]) -> bool:
        """
        Clears the stored Instructions and stores copies of all Instructions in gate_list.

        :param gate_list: the instructions to store
        :return: True if gates were stored correctly, False otherwise (e.g., more gates than $capacity were provided)
        """
        if 0 <= len(gate_list) < self.capacity:
            self.__storage.clear()
            for gate in gate_list:
                self.__storage.append(gate.copy())
            return True
        return False

    def copy_gates(self) -> List[Instruction]:
        """
        Creates a new List that contains copies of all the stored gates.

        :return: a deep copy of the stored gates
        """
        data = []
        for gate in self.__storage:
            data.append(gate.copy())
        return data  # [gate.copy() for gate in self.__storage]

    def instruction_iterator(self) -> Iterator[Instruction]:
        """

        :return: an Iterator over the stored Instructions
        """
        return iter(self.__storage)


class Robot(Controllable, ABC):
    @staticmethod
    def __counts_to_bit_list(counts):  # todo can be deleted I think
        counts = str(counts)
        counts = counts[1:len(counts) - 1]
        arr = counts.split(':')
        if int(arr[1][1:]) != 1:
            Logger.instance().throw(ValueError(f"Function only works for counts with 1 shot but counts was: {counts}"))
        bits = arr[0]
        bits = bits[1:len(bits) - 1]
        list_ = []
        for b in bits:
            list_.append(int(b))
        list_.reverse()  # so that list_[i] corresponds to the measured value of qi
        return list_

    def __init__(self, name: str, attributes: _Attributes, backpack: _Backpack, game_over_callback: Callable[[], None]):
        """

        :param name: name of the Robot for identification and rendering
        :param attributes: describes certain attributes of the Robot
        :param backpack: stores Collectibles
        :param game_over_callback: for stopping the game if the Robot dies
        """
        super().__init__(name)
        self.__score: int = 0
        self.__attributes: _Attributes = attributes
        self.__backpack: _Backpack = backpack
        self.__game_over: Callable[[], None] = game_over_callback
        # initialize qubit stuff (rows)
        self.__simulator = QuantumSimulator()
        self.__unitary_simulator = UnitarySimulator()
        self.__stv: Optional[StateVector] = None
        self.__circuit_matrix: Optional[CircuitMatrix] = None
        self.__qubit_indices: List[int] = []
        for i in range(0, attributes.num_of_qubits):
            self.__qubit_indices.append(i)

        # initialize gate stuff (columns)
        self.__instruction_count: int = 0  # how many instructions are currently placed on the circuit
        # initialize based on empty circuit
        self.__instructions: List[Optional[Instruction]] = [None] * attributes.circuit_space
        # initially there is no static gate (i.e., a gate that cannot be moved and was added by a puzzle)

        if False:
            # todo: for whatever reason this code is slower than calling below method which executes a simulation...
            # todo: (measured execution time is faster, but something is slowed down because the startup time of levels increases)
            self.__stv = StateVector.create_zero_state_vector(self.num_of_qubits)
            self.__circuit_matrix = CircuitMatrix.create_identity(self.num_of_qubits)
        else:
            self.update_statevector(None, use_energy=False, check_for_game_over=False)

    @property
    def used_capacity(self) -> int:
        """
        :return: how many instructions are placed on the robot's circuit
        """
        return self.__backpack.used_capacity

    @property
    def capacity(self) -> int:
        """
        :return: how many instructions can be placed on the robot's circuit
        """
        return self.__backpack.capacity

    @property
    def instructions(self) -> Iterator[Instruction]:
        return self.__backpack.instruction_iterator()

    @property
    def state_vector(self) -> StateVector:
        return self.__stv

    @property
    def circuit_matrix(self) -> CircuitMatrix:
        return self.__circuit_matrix

    @property
    def score(self) -> int:
        return self.__score

    @property
    def cur_energy(self) -> int:
        return self.__attributes.cur_energy

    @property
    def max_energy(self) -> int:
        return self.__attributes.max_energy

    @property
    def num_of_qubits(self) -> int:
        return self.__attributes.num_of_qubits

    @property
    def has_empty_circuit(self) -> bool:
        return self.__instruction_count == 0

    @property
    def circuit_space(self) -> int:
        return self.__attributes.circuit_space

    @property
    def is_space_left(self) -> bool:
        """

        :return: True if the backpack has still space for new Instructions, False if it's already full
        """
        return self.__instruction_count < self.circuit_space

    def game_over_check(self) -> bool:
        """
        Checks if the Robot still has some energy and then calls game_over() if there is None left.

        :return: True if the Robot is game over, False otherwise
        """
        if self.cur_energy <= 0:
            self.__game_over()
            return True
        return False

    def increase_score(self, amount: int):
        if amount < 0:
            Logger.instance().error(f"Tried to increase score by a negative amount: {amount}!", show=False,
                                    from_pycui=False)
            return
        self.__score += amount

    def reset_score(self):
        self.__score = 0

    def key_count(self) -> int:  # cannot be a property since it is an abstractmethod in Controllable
        return self.__backpack.key_count

    def use_key(self) -> bool:
        return self.__backpack.use_key()

    def __update_circuit_space(self, new_circuit_space: int):
        old_instructions = self.__instructions.copy()

        self.__instructions: List[Optional[Instruction]] = [None] * new_circuit_space
        self.__attributes.update_circuit_space(new_circuit_space)

        # copy all placed instructions
        for i, inst in enumerate(old_instructions):
            if inst is not None and i < len(self.__instructions):
                self.__instructions[i] = inst

    def update_statevector(self, input_stv: StateVector, use_energy: bool = True, check_for_game_over: bool = True):
        """
        Compiles and simulates the current circuit and saves and returns the resulting StateVector. Can also lead to a
        game over.

        :param input_stv: custom StateVector used as input
        :param use_energy: whether the update should cost energy or not, defaults to True
        :param check_for_game_over: whether we should perform a game over check or not
        :return: None
        """
        if check_for_game_over and self.game_over_check():
            return

        num_of_used_gates: int = 0  # cannot use len(instructions) since this contains None values
        circuit = QuantumCircuit.from_bit_num(self.num_of_qubits, self.num_of_qubits)
        for inst in self.__instructions:
            if inst is not None:
                num_of_used_gates += 1
                inst.append_to(circuit)

        amplitudes = self.__unitary_simulator.execute(circuit, decimals=QuantumSimulationConfig.DECIMALS)
        self.__circuit_matrix = CircuitMatrix(amplitudes, num_of_used_gates)

        if input_stv is None:  # todo: input_stv might only be None if the circuit is empty (reset or initialized)
            amplitudes = self.__simulator.run(circuit, do_transpile=True)
            self.__stv = StateVector(amplitudes, num_of_used_gates=self.__instruction_count)
        else:
            self.__stv = self.__circuit_matrix.multiply(input_stv)
            if self.__stv is None:
                self.__stv = StateVector.create_zero_state_vector(self.num_of_qubits)

        if use_energy and GameplayConfig.get_option_value(Options.energy_mode):
            self.decrease_energy(amount=1)

    def __remove_instruction(self, instruction: Optional[Instruction], skip_qargs: bool = False) -> bool:
        """
        Tries to remove the given instruction from the circuit. Fails if the Instruction is not used in the circuit.

        :param instruction: the Instruction to remove
        :param skip_qargs: whether to skip resetting the qubits of the Instruction or not, defaults to False
        :return: True if we successfully removed the Instruction, False otherwise
        """
        # todo check if we can extend the condition with "and instruction in self.__instructions"
        if instruction and instruction.is_used():
            self.__instructions[instruction.position] = None
            self.__instruction_count -= 1
            instruction.reset(skip_qargs=skip_qargs)
            return True
        return False

    def remove_instruction(self, instruction: Instruction) -> bool:
        """
        Tries to remove the given instruction from the circuit.

        :param instruction: the Instruction to remove
        :return: True if we successfully removed the Instruction, False otherwise
        """
        if instruction in self.__instructions:
            return self.__remove_instruction(instruction)
        return False

    def __place_instruction(self, instruction: Instruction, position: int) -> bool:
        """
        Tries to place the given instruction at the given position.

        :param instruction: the Instruction we want to place
        :param position: the position in the circuit where we want to place instruction at
        :return: True if instruction was successfully placed at position, False otherwise
        """
        if instruction.position == position:
            return True  # nothing to do in this case
        if instruction.is_used():
            Logger.instance().throw(RuntimeError("Illegal state: Instruction was not removed before placing!"))
            return False

        if 0 <= position < self.circuit_space:
            if self.__instructions[position]:
                self.__remove_instruction(self.__instructions[position])
            self.__instruction_count += 1
            self.__instructions[position] = instruction
            return instruction.use(position)
        else:
            # illegal position removes the instruction from the circuit if possible
            self.__remove_instruction(instruction)
            return False

    def __move_instruction(self, instruction: Instruction, position: int) -> bool:
        """
        Tries to move instruction to position. Fails if instruction is not used or already at position. If an invalid
        position is given, instruction is removed instead.

        :param instruction: the Instruction we want to move
        :param position: the position to which we want to move instruction (invalid position => remove)
        :return: True if we successfully (re)moved instruction, False otherwise
        """
        if instruction.is_used() and instruction.position != position:
            if 0 <= position < self.circuit_space:
                if self.__instructions[position]:
                    self.__remove_instruction(instruction, skip_qargs=True)
                    self.__remove_instruction(self.__instructions[position])
                else:
                    self.__remove_instruction(instruction, skip_qargs=True)
                return self.__place_instruction(instruction, position)
            else:
                self.__remove_instruction(instruction)
                return True
        return False

    def get_stored_instruction(self, index: int) -> Optional[Instruction]:
        """
        Returns the Instruction stored at the given index in the backpack or None for invalid indices.

        :param index: index of the Instruction in the backpack
        :return: the stored Instruction at the given index or None
        """
        if 0 <= index < self.__backpack.used_capacity:
            return self.__backpack.get(index)
        return None

    def use_instruction(self, instruction: Instruction, position: int) -> bool:
        """
        Tries to put the Instruction corresponding to the given index in the backpack into the robot's circuit.
        If the Instruction is already in-use (put onto the circuit) it is removed instead.

        :param instruction: the Instruction we want to use
        :param position: the position of the Instruction in the circuit
        :return: True if we were able to use the Instruction in our circuit
        """
        if instruction.is_used():
            self.__move_instruction(instruction, position)
        else:
            if self.is_space_left or GameplayConfig.get_option_value(Options.allow_implicit_removal):
                return self.__place_instruction(instruction, position)
            else:
                return False
        return True

    def reset_circuit(self):
        """
        Resets the circuit by removing all Instructions of it and updating the statevector.

        :return: None
        """
        temp = self.__instructions.copy()
        for instruction in temp:
            self.__remove_instruction(instruction)
        self.__instruction_count = 0
        self.__stv = StateVector.create_zero_state_vector(self.num_of_qubits)
        self.__circuit_matrix = CircuitMatrix.create_identity(self.num_of_qubits)

    def get_available_instructions(self) -> List[Instruction]:
        """

        :return: a List containing copies of all Instructions currently available to this Robot
        """
        return self.__backpack.copy_gates()

    def set_available_instructions(self, gates: Union[List[Instruction], List[GateType]]) -> bool:
        """
        :param gates: either a list of GateType or Instruction that determines which Instructions are available
        :return: True if gates were successfully set, False if an error occurred
        """
        if len(gates) > 0 and isinstance(gates[0], GateType):
            gate_types: List[GateType] = gates
            gates = [InstructionManager.from_type(gt) for gt in gate_types]
        return self.__backpack.set_instructions(gates)

    def give_collectible(self, collectible: Collectible, force: bool = False) -> bool:
        """
        Gives collectible to this Robot.

        :param collectible: the Collectible we want to give this Robot
        :param force: whether we force collectible to be added, defaults to False
        :return: True if the collectible was given successfully, False otherwise
        """
        if isinstance(collectible, Score):
            self.__score += collectible.amount
            return True
        elif isinstance(collectible, Key):
            return self.__backpack.give_key(collectible.amount)
        elif isinstance(collectible, Energy):
            self.__attributes.increase_energy(collectible.amount)
            return True
        elif isinstance(collectible, Instruction):
            return self.__backpack.add(collectible, force)
        elif isinstance(collectible, Qubit):
            self.__attributes.add_qubits(collectible.additional_qubits)
            return True
        elif isinstance(collectible, MultiCollectible):
            success = True
            for c in collectible.iterator():
                success = success and self.give_collectible(c, force)
            return success
        else:
            Logger.instance().error(f"Received uncovered collectible: {collectible}", show=False, from_pycui=False)
            return False

    def on_move(self):
        """
        Does nothing now!
        No longer decreases energy every time the Robot moves. This feature turned out to be tedious and doesn't
        support the game's goal of making you familiar with Quantum Computing in any way.

        :return: None
        """
        # self.__attributes.decrease_energy(amount=1)
        pass

    def decrease_energy(self, amount: int = 1) -> Tuple[int, bool]:
        """
        Decreases this Robot's current energy by the given amount.

        :param amount: by how much we want to reduce this Robot's energy, defaults to 1
        :return: the actual amount by how much current energy was decreased, whether this Robot is game over or not
        """
        assert amount > 0  # todo maybe == 0 is also okay?

        if self.game_over_check():
            return amount, True
        return self.__attributes.decrease_energy(amount), False

    def increase_energy(self, amount: int = 1) -> int:
        """
        Increases this Robot's current energy by the given amount at most to its max energy.

        :param amount: by how much current energy was increased
        :return: by how much current energy was actually increased.
        """
        assert amount > 0  # todo maybe == 0 is also okay?
        return self.__attributes.increase_energy(amount)

    def gate_used_at(self, position: int) -> Optional[Instruction]:
        """
        Returns the Instruction at the given index or None for invalid indices.

        :param position: position of the Instruction in the circuit
        :return: the Instruction at the given position or None
        """
        if 0 <= position < self.circuit_space:
            return self.__instructions[position]
        return None

    def reset(self):
        self.__attributes.increase_energy(self.__attributes.max_energy)
        self.reset_circuit()


class BaseBot(Robot):
    @staticmethod
    def from_properties(properties: RoboProperties, game_over_callback: Callable[[], None]) -> "BaseBot":
        return BaseBot(game_over_callback, num_of_qubits=properties.num_of_qubits,
                       circuit_space=properties.circuit_space, gates=properties.instruction_list)

    def __init__(self, game_over_callback: Callable[[], None], num_of_qubits: int = 3,
                 gates: Optional[List[Instruction]] = None,
                 circuit_space: Optional[int] = None, backpack_space: Optional[int] = None,
                 max_energy: Optional[int] = None, start_energy: Optional[int] = None):
        attributes = _Attributes(DummyQubitSet(num_of_qubits), circuit_space, max_energy, start_energy)
        backpack = _Backpack(backpack_space, gates)
        super(BaseBot, self).__init__("BaseBot", attributes, backpack, game_over_callback)

    def get_img(self):
        return "Q"

    def description(self) -> str:
        return "The most basic Robot"


class LukeBot(Robot):
    def __init__(self, game_over_callback: Callable[[], None], size: int = 2):
        attributes = _Attributes(DummyQubitSet(size))
        backpack = _Backpack(capacity=5)

        # randomness is not allowed during Robot creation because it messes up the seed
        # add random gates and a HealthPotion
        # rm = RandomManager.create_new()
        # if rm.get(msg="LukeBot.init()") < 0.5:
        #     num_of_gates = 3
        # else:
        #     num_of_gates = 4
        # gate_factory = GateFactory.default()
        # for gate in gate_factory.produce_multiple(rm, num_of_gates):
        #     backpack.add(gate)
        # backpack.place_in_pouch(HealthPotion(3))
        super().__init__("Luke", attributes, backpack, game_over_callback)

    def get_img(self):
        return "L"

    def description(self) -> str:
        return "The loyal Robot you start with. A true all-rounder - take good care of it!"


class FusionBot(Robot):
    def __init__(self, num_of_qubits: int, gates: List[Instruction]):
        attributes = _Attributes(DummyQubitSet(num_of_qubits), len(gates))
        backpack = _Backpack(len(gates), gates)
        super().__init__("FusionBot", attributes, backpack, lambda: None)

    def get_img(self):
        return "F"

    def description(self) -> str:
        return "A Robot used to fuse gates."



from enum import Enum
from typing import Dict, Optional, List, Callable, Tuple, Any

from .test_config import TestConfig


class OptionsType(Enum):
    Bool = 0
    Float = 1

    def get_callbacks(self, data: Any) -> Tuple[Callable[[int], str], Callable[[str], Any]]:
        if self is OptionsType.Bool:
            return OptionsType._get_boolean_callback()
        elif self is OptionsType.Float:
            if isinstance(data, Tuple) and len(data) == 3:
                min_, max_, steps = data
                if max_ <= min_:
                    raise Exception(f"OptionsType.Float's data[0] must be smaller than data[1], but got \"{data}\"")
                if steps < 0:
                    raise Exception(f"OptionsType.Float's data[2] must be positive, but got \"{data}\"")
                return OptionsType._get_float_callback(min_, max_, steps)
            raise Exception(f"Expected Tuple of length=3 as OptionsType.Float's data but got \"{data}\".")
        raise Exception(f"Callbacks for OptionsType \"{self.name}\" not yet implemented!")

    @staticmethod
    def _get_boolean_callback() -> Tuple[Callable[[int], str], Callable[[str], bool]]:
        def get(index: int) -> str:
            if index % 2 == 1:
                return "yes"
            else:
                return "no"

        def convert(value: str) -> bool:
            if value.lower() == "yes":
                return True
            else:
                return False

        return get, convert

    @staticmethod
    def _get_float_callback(min_: float, max_: float, steps: int) -> Tuple[Callable[[int], str], Callable[[str], float]]:
        assert min_ < max_
        assert steps > 0
        range_ = max_ - min_
        step_size = range_ / steps

        def get(index: int) -> str:
            val = round(min_ + step_size * index, ndigits=3)
            return str(val)

        return get, float


class Options(Enum):
    energy_mode = ("Energy Mode", OptionsType.Bool, None, 2, 0,
                   "In energy mode every change to your circuit costs energy. You loose if your Robot is out of "
                   "energy.")

    auto_save = ("Auto Save", OptionsType.Bool, None, 2, 1,
                 "Whether to automatically save the game on exit or not.")
    auto_reset_circuit = ("Auto Reset Circuit", OptionsType.Bool, None, 2, 1,
                          "Automatically reset your Circuit to a clean state at the beginning of a Puzzle, Riddle, "
                          "etc.")
    log_keys = ("Log Keys", OptionsType.Bool, None, 2, 1,
                "Stores all keys you pressed in a .qrkl-file so one can replay them (e.g. for analysing a bug)")

    gameplay_key_pause = ("Gameplay Key Pause", OptionsType.Float, (0.1, 1.0, 9), 9, 0,
                          "How long to wait before we process the next input during gameplay.")
    simulation_key_pause = ("Simulation Key Pause", OptionsType.Float, (0.05, 1.0, 19), 19, 3,
                            "How long to wait before we process the next input during simulation.")

    show_ket_notation = ("Show Ket-Notation", OptionsType.Bool, None, 2, 1,
                         "Whether to display ket-notation for state vectors and circuit matrix or not.")
    allow_implicit_removal = ("Allow implicit Removal", OptionsType.Bool, None, 2, 0,
                              "Allows you to place a gate on an occupied spot, removing the occupying gate in the "
                              "process.")
    allow_multi_move = ("Allow multi move", OptionsType.Bool, None, 2, 0,
                        "Allows you to move multiple tiles at once by pressing a number followed by a direction.")
    auto_skip_text_animation = ("Auto skip text animation", OptionsType.Bool, None, 2, 0,
                                "During some special scene transitions there will be some animated text describing "
                                "what happened in-between story sections. Skipping this means that the whole text "
                                "will be shown at once.")
    enable_puzzle_history = ("Enable puzzle history", OptionsType.Bool, None, 2, 1,
                             "Whether you can navigate through your circuit's history or not while solving a puzzle.")
    auto_reset_history = ("Auto reset history", OptionsType.Bool, None, 2, 1,
                          "Whether the puzzle history should automatically jump to the present (= last changes) when "
                          "you navigate through menus.\nNote: Confirming to edit the circuit always resets history.")

    def __init__(self, name: str, otype: OptionsType, data: Any, num_of_values: int, default_index: int,
                 description: str):
        self.__name = name
        self.__otype = otype
        self.__data = data
        self.__get_value, self.__convert_value = otype.get_callbacks(data)
        self.__num_of_values = num_of_values
        self.__default_index = default_index
        self.__description = description

    @property
    def name(self) -> str:
        return self.__name

    @property
    def type(self) -> OptionsType:
        return self.__otype

    @property
    def data(self) -> Any:
        return self.__data

    @property
    def num_of_values(self) -> int:
        return self.__num_of_values

    @property
    def default_index(self) -> int:
        return self.__default_index

    @property
    def description(self) -> str:
        return self.__description

    def get_value(self, cur_index: int) -> str:
        return self.__get_value(cur_index)

    def convert(self, cur_index: int) -> Any:
        value = self.get_value(cur_index)
        return self.__convert_value(value)


class OptionsManager:
    __SEPARATOR = "="
    __OPTIONS: Dict[Options, int] = {}

    @staticmethod
    def validate() -> bool:
        # todo: check conversion of value of random index for options?
        try:
            for option in Options:
                option.get_value(0)
            return True
        except:
            return False

    @staticmethod
    def init():
        for val in Options:
            OptionsManager.__OPTIONS[val] = val.default_index

    @staticmethod
    def get_options_next_callback(option: Options) -> Callable[[], str]:
        def next_() -> str:
            # first increment the current index
            next_index = OptionsManager.__OPTIONS[option] + 1
            if next_index >= option.num_of_values:
                next_index = 0
            OptionsManager.__OPTIONS[option] = next_index
            # then return the corresponding new value
            return option.get_value(next_index)
        return next_

    @staticmethod
    def get_option_value(option: Options, convert: Optional[bool] = None, ignore_test_config: Optional[bool] = None) \
            -> Any:
        if convert is None: convert = True
        if ignore_test_config is None: ignore_test_config = False

        cur_index = OptionsManager.__OPTIONS[option]
        if convert:
            if option is Options.simulation_key_pause and TestConfig.is_active() and not ignore_test_config:
                return TestConfig.key_pause()
            return option.convert(cur_index)
        else:
            return option.get_value(cur_index)

    @staticmethod
    def set_option_value(option: Options, value: Any) -> bool:
        try:
            index = OptionsManager._get_closest_index(option, value)
            OptionsManager.__OPTIONS[option] = index
            return True
        except:
            return False

    @staticmethod
    def to_string() -> str:
        lines = [f"{option.name}{OptionsManager.__SEPARATOR}{option.convert(OptionsManager.__OPTIONS[option])}"
                 for option in OptionsManager.__OPTIONS]
        return "\n".join(lines)

    @staticmethod
    def _get_closest_index(option: Options, value: Any) -> int:
        if option.type is OptionsType.Bool:
            if isinstance(value, str):
                return 1 if value.lower().strip() in ["true", "yes", "1"] else 0
            else:
                return 1 if value == 1 else 0

        elif option.type is OptionsType.Float:
            if isinstance(value, str):
                value = float(value.strip())
            if isinstance(value, float):
                min_, max_, steps = option.data
                if value <= min_: return 0
                if value >= max_: return steps  # steps is the maximum index
                # remove the offset of min_ and then check how often step_diff is in the remaining value
                step_diff = (max_ - min_) / steps
                return int(round((value - min_) / step_diff))
            else:
                # invalid type for value, so we return the default index
                return option.default_index

        raise NotImplementedError(f"OptionsType {option.type} not implemented!")

    @staticmethod
    def from_text(log_text: str) -> bool:
        def normalize(text: str) -> str:
            return text.lower().strip(" ")

        for line in log_text.splitlines():
            if len(line.strip(" ")) == 0:
                continue
            try:
                name, value = line.split(OptionsManager.__SEPARATOR)
                name = normalize(name)
                for option in Options:
                    if normalize(option.name) == name:  # compare the normalized names
                        index = OptionsManager._get_closest_index(option, value)
                        OptionsManager.__OPTIONS[option] = index
                        break
            except IndexError:
                return False
            except KeyError:
                return False
        return True



from abc import ABC, abstractmethod
from enum import Enum
from typing import Iterator, List, Callable, Optional


class CollectibleType(Enum):
    # Consumable = 1      # currently unused
    ActiveItem = 3  # currently unused
    PassiveItem = 4  # currently unused

    Multi = 0  # wraps multiple collectibles
    Gate = 2
    Qubit = 6

    Pickup = 5  # for undefined pickups
    Key = 51
    # Coin = 52
    Energy = 53
    Score = 54

    QuantumFuser = 70


class Collectible(ABC):
    def __init__(self, c_type: CollectibleType):
        self.__type = c_type

    @property
    def type(self):
        return self.__type

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        pass

    @abstractmethod
    def to_string(self) -> str:
        pass


class MultiCollectible(Collectible):
    PRICE_MULT = 0.9

    def __init__(self, content: List[Collectible]):
        super(MultiCollectible, self).__init__(CollectibleType.Multi)
        self.__content = content

    def name(self) -> str:
        return "Collectible Pack"

    def description(self, check_unlocks: Optional[Callable[[str], bool]] = None) -> str:
        desc = "Contains multiple Collectibles:"
        for collectible in self.__content:
            desc += "\n  - " + collectible.name()
        return desc

    def to_string(self) -> str:
        text = "Multi ["
        for collectible in self.__content:
            text += collectible.to_string() + ", "
        return text + "]"

    def iterator(self) -> Iterator[Collectible]:
        return iter(self.__content)



from datetime import datetime
from typing import Tuple, Optional, List, Union, Dict

from antlr4 import InputStream, CommonTokenStream

from qrogue.game.logic.collectibles import InstructionManager, Instruction
from qrogue.game.logic.collectibles.instruction import CombinedGate
from qrogue.graphics.popups import Popup
from qrogue.management.save_grammar.SaveDataLexer import SaveDataLexer
from qrogue.management.save_grammar.SaveDataParser import SaveDataParser
from qrogue.management.save_grammar.SaveDataVisitor import SaveDataVisitor
from qrogue.util import Logger, CommonInfos, LevelInfo, LevelData, Config, PathConfig, FileTypes, ParserErrorListener, \
    GateType, GameplayConfig
from qrogue.util.achievements import Achievement, Unlocks
from qrogue.util.util_functions import cur_datetime, datetime2str, simple_decode, simple_encode


class NewSaveData:
    __instance = None
    __ENCODING_KEY = "6l5f5BqbldikHYpqE240"

    @staticmethod
    def _encode(data: str, force: bool = False) -> str:
        if not Config.debugging() or force:
            data = simple_encode(NewSaveData.__ENCODING_KEY, data)
        return data

    @staticmethod
    def _decode(data: str) -> str:
        if data.startswith(_SaveDataGenerator.header()):
            return data
        return simple_decode(NewSaveData.__ENCODING_KEY, data)

    @staticmethod
    def load(path: str, in_user_path: bool = True) -> "NewSaveData":
        if not path.endswith(FileTypes.Save.value):
            path += FileTypes.Save.value
        try:
            save_data = NewSaveData._decode(PathConfig.read(path, in_user_path))
            return NewSaveData(save_data)

        except FileNotFoundError as ex:
            Popup.error(f"Could not load save file at \"{path}\": {ex}.\nPlease make sure that you did not move your "
                        f"files manually. If this error occurs even though the stated path points to a "
                        f"{FileTypes.Save.value}-file:", add_report_note=True)

    @staticmethod
    def empty_save_state() -> str:
        save_state = f"{_SaveDataGenerator.header()}\n"
        save_state += f"{datetime2str(cur_datetime())}\n"
        save_state += f"{_SaveDataGenerator.inventory_header()}\n"
        save_state += f"{_SaveDataGenerator.quantum_fuser()} 0\n"
        save_state += f"{_SaveDataGenerator.gates_header()}\n"
        save_state += f"{_SaveDataGenerator.levels_header()}\n"
        save_state += f"{_SaveDataGenerator.unlocks_header()}\n"
        save_state += f"{_SaveDataGenerator.achievements_header()}\n"
        save_state += f"{_SaveDataGenerator.ender()}\n"
        return save_state

    @staticmethod
    def _init_fresh_save(save_data: "NewSaveData"):
        # some achievements need to be present right from the start
        expeditions = Achievement(Achievement.CompletedExpedition, 0, 100)
        save_data.__achievements[expeditions.name] = expeditions

        # some unlocks are present by default for backwards compatibility   # todo: fix this because it's confusing on the Achievement-screen
        save_data.unlock(Unlocks.GateRemove)

    def __init__(self, save_data: Optional[str] = None):
        """

        Args:
            save_data: if None is provided, the data is loaded from the latest save file
        """

        if save_data is None:
            path = PathConfig.find_latest_save_file()
            # a fresh save has no digit before the file ending
            self.__is_fresh_save = not path[-len(FileTypes.Save.value) - 1].isdigit()
            if not self.__is_fresh_save:
                try:
                    save_data = PathConfig.read(path, in_user_path=True)
                except FileNotFoundError:
                    Logger.instance().throw(NotImplementedError("This line should not be reachable! Please send us the "
                                                                "log files so we can fix the issue as soon as "
                                                                "possible. Thank you!"))
        else:
            self.__is_fresh_save = False

        self.__date_time = cur_datetime()  # date and time of the latest save
        self.__inventory = _SaveDataGenerator.Inventory.default()
        self.__gates: List[Instruction] = []
        self.__levels: Dict[str, LevelData] = {}  # key is a level's internal name
        self.__achievements: Dict[str, Achievement] = {}
        self.__unlocks: Dict[str, datetime] = {}
        self.__has_unsaved_changes = False

        if save_data is not None and len(save_data.strip()) > 0:
            save_data = self._decode(save_data)
            generator = _SaveDataGenerator()
            self.__date_time, self.__inventory, gates, levels, unlocks, achievement_list = generator.load(save_data)
            self.__gates = gates.copy()
            for level in levels: self.__levels[level.name] = level
            for ach in achievement_list: self.__achievements[ach.name] = ach
            for name, date_time2 in unlocks: self.__unlocks[name] = date_time2

        if self.__is_fresh_save:
            NewSaveData._init_fresh_save(self)

    @property
    def is_fresh_save(self) -> bool:
        return self.__is_fresh_save

    @property
    def num_quantum_fusers(self) -> int:
        return self.__inventory.quantum_fuser

    def check_level(self, internal_name: str) -> bool:
        return internal_name in self.__levels

    def check_unlocks(self, unlock: Union[str, Unlocks]) -> bool:
        if isinstance(unlock, Unlocks): unlock = unlock.ach_name
        return unlock in self.__unlocks

    def check_achievement(self, name: str) -> bool:
        return name in self.__achievements and self.__achievements[name].is_done()

    def check(self, name: str) -> bool:
        if self.check_level(name): return True
        if self.check_unlocks(name): return True
        if self.check_achievement(name): return True
        return False

    def _complete_map(self, name: str, duration: int, date_time: Optional[datetime] = None, score: int = -1) \
            -> LevelData:
        # NOTE: name might still have the "done" suffix! Use level_data.name if you need the normalized name.
        # compute date_time and duration based on current time if not provided
        if date_time is None: date_time = cur_datetime()

        level_data = LevelData(name, date_time, duration, score)
        if level_data.name in self.__levels and level_data.total_score >= self.__levels[level_data.name].total_score:
            # save the better score for the replayed level
            self.__levels[level_data.name] = level_data
            self.__has_unsaved_changes = True

        elif level_data.name not in self.__levels:
            # save the level because it has never been completed before
            self.__levels[level_data.name] = level_data
            self.__has_unsaved_changes = True

            # add potential unlocks corresponding to the level
            for unlock in LevelInfo.get_level_completion_unlocks(level_data.name, self.check_level):
                self.unlock(unlock, date_time)

            # save the Instructions corresponding to the gate types that were unlocked
            for gate_type in LevelInfo.get_level_completion_unlocked_gates(level_data.name, self.check_level):
                self.__gates.append(InstructionManager.from_type(gate_type))

        return level_data

    def complete_level(self, name: str, duration: int, date_time: Optional[datetime] = None, score: int = -1) \
            -> LevelData:
        return self._complete_map(name, duration, date_time, score)

    def complete_expedition(self, name: str, duration: int, difficulty_level: int, gate_type: GateType,
                            date_time: Optional[datetime] = None, score: int = -1) \
            -> LevelData:
        level_data = self._complete_map(name, duration, date_time, score)
        self.add_to_achievement(Achievement.CompletedExpedition, difficulty_level)
        self.__gates.append(InstructionManager.from_type(gate_type))
        self.__inventory.quantum_fuser += 1     # todo: will an expedition's Boss always give 1 QuantumFuser?
        return level_data

    def unlock(self, unlock: Union[str, Unlocks], date_time: Optional[datetime] = None):
        if date_time is None: date_time = cur_datetime()
        if isinstance(unlock, Unlocks): unlock = unlock.ach_name
        if unlock in self.__unlocks:
            return
        self.__unlocks[unlock] = date_time
        self.__has_unsaved_changes = True

    def get_level_data(self, level_name: str) -> Optional[LevelData]:
        if level_name in self.__levels:
            return self.__levels[level_name]
        return None

    def get_progress(self, achievement: str) -> Tuple[float, float]:
        if achievement in self.__achievements:
            ach = self.__achievements[achievement]
            return ach.score, ach.done_score
        return -1, -1

    def get_gates(self) -> List[Instruction]:
        return [gate.copy() for gate in self.__gates]

    def get_original_gates(self) -> List[Instruction]:
        return self.__gates.copy()

    def decompose(self, gates: Union[Instruction, List[Instruction]]) -> bool:
        if isinstance(gates, Instruction):
            if gates in self.__gates:
                self.__gates.remove(gates)
                self.__inventory.quantum_fuser += 1     # decomposing a gate gives 1 QuantumFuser
                self.__has_unsaved_changes = True
                return True
        else:
            # check if all gates we want to decompose are indeed among our gates
            if False not in [gate in self.__gates for gate in gates]:
                for gate in gates: self.__gates.remove(gate)
                self.__inventory.quantum_fuser += len(gates)
                self.__has_unsaved_changes = True
                return True
        return False

    def add_gate(self, gate: Instruction):
        # copy and reset the new gate
        gate = gate.copy()
        gate.reset()
        self.__gates.append(gate)
        self.__has_unsaved_changes = True

    def to_achievements_string(self) -> str:
        # todo: improve readability
        text = ""
        if len(self.__levels) > 0:
            text += f"-- Completed Levels --\n"
            text += "\n".join([f"{level.name} @ {datetime2str(level.date_time)} "
                               f"{level.duration} {_SaveDataGenerator.duration_unit()} Score = {level.score}"
                               for level in self.__levels.values()])
            text += "\n\n"

        if len(self.__unlocks) > 0:
            text += f"-- Unlocks --\n"
            text += "\n".join([f"{unlock} @ {datetime2str(self.__unlocks[unlock])}"
                               for unlock in self.__unlocks])
            text += "\n\n"

        text += f"-- Achievements --\n"
        text += "\n".join([f"{ach.name} @ {datetime2str(ach.date_time)} "
                           f"Score = {ach.score} out of {ach.done_score}"
                           for ach in self.__achievements.values()])
        return text

    def to_keylog_string(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        text = f"{_SaveDataGenerator.header()}\n"
        text += f"{datetime2str(self.__date_time)}\n"

        text += f"{_SaveDataGenerator.inventory_header()}\n"
        text += f"{self.__inventory.to_string()}\n"

        text += f"{_SaveDataGenerator.gates_header()}\n"
        if len(self.__gates) > 0:
            text += _SaveDataGenerator.gate_separator().join([gate.to_save_string() for gate in self.__gates])
            text += "\n"

        text += f"{_SaveDataGenerator.levels_header()}\n"
        for level in self.__levels.values():
            text += f"{level.name} @ {datetime2str(level.date_time)} {level.duration} " \
                    f"{_SaveDataGenerator.duration_unit()} Score = {level.score}\n"

        text += f"{_SaveDataGenerator.unlocks_header()}\n"
        for unlock in self.__unlocks:
            text += f"{unlock} @ {datetime2str(self.__unlocks[unlock])}\n"

        text += f"{_SaveDataGenerator.achievements_header()}\n"
        for ach in self.__achievements.values():
            text += f"{ach.name} @ {datetime2str(ach.date_time)} Score = {ach.score} out of {ach.done_score}\n"

        return text + f"{_SaveDataGenerator.ender()}\n"

    def save(self, is_auto_save: bool = False) -> Tuple[bool, CommonInfos]:
        """
        Returns:
            False if an error occurred during saving, True if saving behaved as expected (i.e., latest save state is
            persisted)
        """
        if Config.forbid_saving():
            return False, CommonInfos.NoSavingWithCheats
        if not self.__has_unsaved_changes:
            return True, CommonInfos.NothingToSave

        try:
            self.__date_time = cur_datetime()  # update datetime of the latest save (=now)
            data = self._encode(self.to_string())
            if is_auto_save:
                PathConfig.write_auto_save(data)
            else:
                PathConfig.new_save_file(data)
                self.__has_unsaved_changes = False  # only change flag if it was a manual (i.e., no auto) save
            return True, CommonInfos.SavingSuccessful

        except Exception as ex:
            Logger.instance().error(f"Exception occurred during saving: {ex}", False, False)
            return False, CommonInfos.SavingFailed

    def compare(self, other: "NewSaveData") -> Tuple[List[Instruction], List[str], List[Unlocks], List[Achievement]]:
        """
        Args:
            other: the NewSaveData to compare with
        Returns:
            List of Instructions, level names, Unlocks and Achievement other has and self does not
        """
        other_gates = other.__gates.copy()
        self_gates = self.__gates.copy()
        gate_diff = []
        # find gates that are in other_gates but not in self_gates
        for gate in other_gates:
            if gate in self_gates:
                # remove gate from self_gates, so we can also check if the number of specific gate_types match
                self_gates.remove(gate)
                continue
            gate_diff.append(gate.copy())   # append a copy so the outside cannot manipulate the actual gate

        level_diff = []
        for level in other.__levels:
            if self.check_level(level): continue
            level_diff.append(level)

        unlocks_diff = []
        for unlock in other.__unlocks:
            if self.check_unlocks(unlock): continue
            unlocks_diff.append(Unlocks.from_name(unlock))

        achievement_diff = []
        for ach_name in other.__achievements:
            if self.check_achievement(ach_name): continue
            achievement_diff.append(other.__achievements[ach_name])

        return gate_diff, level_diff, unlocks_diff, achievement_diff

    def get_completed_levels(self) -> List[LevelData]:
        return list(self.__levels.values())

    def add_to_achievement(self, name: str, score: float = 1):
        if name not in self.__achievements:
            raise Exception("Use add_achievement() for new achievements!")

        self.__achievements[name].add_score(score)
        self.__has_unsaved_changes = True


class _SaveDataGenerator(SaveDataVisitor):
    class Inventory:
        @staticmethod
        def default() -> "_SaveDataGenerator.Inventory":
            return _SaveDataGenerator.Inventory(0)

        def __init__(self, quantum_fuser: int):
            # for every gate we want to fuse, we need to pay one quantum fuser
            # e.g., fusing HGate@q1 with CXGate@q1q0 costs two quantum fuser
            self.quantum_fuser = quantum_fuser

        def to_string(self) -> str:
            return f"{_SaveDataGenerator.quantum_fuser()} {self.quantum_fuser}"

        def __str__(self):
            return f"Inventory({self.quantum_fuser})"

    @staticmethod
    def header() -> str:
        return "Qrogue<"

    @staticmethod
    def inventory_header() -> str:
        return "[INVENTORY]"

    @staticmethod
    def gates_header() -> str:
        return "[GATES]"

    @staticmethod
    def levels_header() -> str:
        return "[LEVELS]"

    @staticmethod
    def unlocks_header() -> str:
        return "[UNLOCKS]"

    @staticmethod
    def achievements_header() -> str:
        return "[ACHIEVEMENTS]"

    @staticmethod
    def ender() -> str:
        return ">Qrogue"

    @staticmethod
    def duration_unit() -> str:
        return "seconds"

    @staticmethod
    def quantum_fuser() -> str:
        return "QuantumFuser"

    @staticmethod
    def gate_separator() -> str:
        return ";"  # the general separator can be used to separate gates for improved readability

    def __init__(self):
        self.__knowledge_mode = None
        self.__highest_knowledge_level = -1

    def load(self, file_data: str) -> Tuple[datetime, Inventory, List[Instruction], List[LevelData],
             List[Tuple[str, datetime]], List[Achievement]]:
        input_stream = InputStream(file_data)
        lexer = SaveDataLexer(input_stream)
        token_stream = CommonTokenStream(lexer)
        parser = SaveDataParser(token_stream)
        parser.addErrorListener(ParserErrorListener())

        return self.visitStart(parser.start())

    #####################################

    def visitDate_time(self, ctx: SaveDataParser.Date_timeContext) -> datetime:
        date: str = ctx.DATE().getText()
        day = int(date[0:2])
        month = int(date[3:5])
        year = int(date[6:10])

        time: str = ctx.TIME().getText()
        hour = int(time[0:2])
        minute = int(time[3:5])
        second = int(time[6:8])

        return datetime(year, month, day, hour, minute, second)

    def visitValue(self, ctx: SaveDataParser.ValueContext) -> int:
        return int(ctx.VALUE().getText())

    def visitDuration(self, ctx: SaveDataParser.DurationContext) -> int:
        return self.visitValue(ctx.value())

    def visitScore(self, ctx: SaveDataParser.ScoreContext) -> int:
        return self.visitValue(ctx.value())

    #####################################

    def visitInventory(self, ctx: SaveDataParser.InventoryContext) -> Inventory:
        quantum_fuser = self.visitValue(ctx.value())
        return _SaveDataGenerator.Inventory(quantum_fuser)

    #####################################

    def visitBackground_gate(self, ctx: SaveDataParser.Background_gateContext) -> Optional[Instruction]:
        gate = InstructionManager.instruction_from_name(ctx.NAME_STD().getText())
        if gate is None: return None

        qubits = [self.visitValue(val) for val in ctx.value()]
        gate.setup(qubits)
        return gate

    def visitCombined_gate(self, ctx: SaveDataParser.Combined_gateContext) -> Optional[Instruction]:
        name = ctx.TEXT().getText()[1:-1]   # remove leading and trailing quotation mark
        num_of_qubits = self.visitValue(ctx.value())

        gate_list = [self.visitBackground_gate(bg_ctx) for bg_ctx in ctx.background_gate()]
        if None in gate_list:
            Logger.instance().error(f"Failed to load a background gate of CombinedGate \"{name}\"")
            return None
        return CombinedGate(gate_list, num_of_qubits, name)

    def visitGate(self, ctx: SaveDataParser.GateContext) -> Optional[Instruction]:
        if ctx.combined_gate() is None:
            return InstructionManager.instruction_from_name(ctx.NAME_STD().getText())
        else:
            return self.visitCombined_gate(ctx.combined_gate())

    def visitGates(self, ctx: SaveDataParser.GatesContext) -> List[Instruction]:
        gates = []
        for gate_ctx in ctx.gate():
            gate = self.visitGate(gate_ctx)
            if gate is not None:
                gates.append(gate)
        return gates

    #####################################

    def visitLevel(self, ctx: SaveDataParser.LevelContext) -> LevelData:
        name = ctx.NAME_SPECIAL().getText()
        date_time = self.visitDate_time(ctx.date_time())
        duration = self.visitDuration(ctx.duration())
        score = self.visitScore(ctx.score())
        level_data = LevelData(name, date_time, duration, score)

        # find out which knowledge mode the highest level has
        km, ln = level_data.knowledge_mode, level_data.level_num
        if self.__knowledge_mode is None:
            # initialize knowledge mode regardless (could still be null if level has no knowledge mode)
            self.__knowledge_mode = km
        elif km is not None and ln is not None:
            # only set it if km and ln are not None (if km is None, we don't have any information anyways, and if ln is
            #  None we cannot compare it to the highest level anyways)
            if km != self.__knowledge_mode and ln is not None and ln > self.__highest_knowledge_level:
                # set new value for knowledge mode if km differs and comes from a higher level
                self.__knowledge_mode = km
            # store new highest level regardless of whether knowledge mode changed (if it did, we already know that ln
            #  is greater than the current highest level, else we stay in the same mode and also want to update highest
            #  level)
            self.__highest_knowledge_level = max(self.__highest_knowledge_level, ln)

        return level_data

    def visitLevels(self, ctx: SaveDataParser.LevelsContext) -> List[LevelData]:
        return [self.visitLevel(level) for level in ctx.level()]

    #####################################
    def visitUnlock(self, ctx: SaveDataParser.UnlockContext) -> Tuple[str, datetime]:
        name = ctx.NAME_STD().getText()
        date_time = self.visitDate_time(ctx.date_time())
        return name, date_time

    def visitUnlocks(self, ctx: SaveDataParser.UnlocksContext) -> List[Tuple[str, datetime]]:
        return [self.visitUnlock(unlock) for unlock in ctx.unlock()]

    #####################################
    def visitAchievement(self, ctx: SaveDataParser.AchievementContext) -> Achievement:
        name = ctx.NAME_STD().getText()
        date_time = self.visitDate_time(ctx.date_time())
        actual_score = self.visitScore(ctx.score())
        required_score = self.visitValue(ctx.value())

        return Achievement(name, actual_score, required_score, date_time)

    def visitAchievements(self, ctx: SaveDataParser.AchievementsContext) -> List[Achievement]:
        return [self.visitAchievement(achievement) for achievement in ctx.achievement()]

    #####################################

    def visitStart(self, ctx: SaveDataParser.StartContext) -> Tuple[datetime, Inventory, List[Instruction],
            List[LevelData], List[Tuple[str, datetime]], List[Achievement]]:
        date_time = self.visitDate_time(ctx.date_time())

        if ctx.inventory():
            inventory = self.visitInventory(ctx.inventory())
        else:
            inventory = _SaveDataGenerator.Inventory.default()

        gates = self.visitGates(ctx.gates())
        levels = self.visitLevels(ctx.levels())
        unlocks = self.visitUnlocks(ctx.unlocks())
        achievement_list = self.visitAchievements(ctx.achievements())

        if not GameplayConfig.set_knowledge_mode(self.__knowledge_mode):
            Logger.instance().warn(f"Failed to set knowledge mode to \"{self.__knowledge_mode}\"", from_pycui=False)

        return date_time, inventory, gates, levels, unlocks, achievement_list



from typing import List, Union, Tuple, Optional

from qrogue.game.logic.actors.controllables import Robot
from qrogue.game.logic.base import StateVector
from qrogue.game.logic.collectibles import Collectible, CollectibleFactory, Instruction
from qrogue.util import Logger, MyRandom


class ExplicitTargetDifficulty:
    """
    A TargetDifficulty that doesn't create StateVectors based on a Robot's possibilities but by choosing from a pool
    of explicitly provided StateVectors
    """

    def __init__(self, stv_pool: List[StateVector], reward: Optional[Union[CollectibleFactory, Collectible]] = None,
                 ordered: bool = False):
        """

        :param stv_pool: list of StateVectors to choose from
        :param reward: factory for creating a reward or a specific reward (Collectible)
        :param ordered: whether StateVectors should be chosen in order or randomly from the given stv_pool
        """
        self.__pool = stv_pool
        self.__ordered = ordered
        self.__order_index = -1
        if reward is None:
            self.__reward_factory = None
        elif isinstance(reward, CollectibleFactory):
            self.__reward_factory = reward
        else:
            self.__reward_factory = CollectibleFactory([reward])

    @property
    def has_reward_factory(self) -> bool:
        return self.__reward_factory is not None

    def create_statevector(self, robot: Robot, rm: MyRandom) -> StateVector:
        if self.__ordered or rm is None:
            self.__order_index += 1
            if self.__order_index >= len(self.__pool):
                self.__order_index = 0
            stv = self.__pool[self.__order_index]
        else:
            stv = rm.get_element(self.__pool, msg="ExplicitTargetDiff_selectStv")

        if stv.num_of_qubits != robot.num_of_qubits:
            Logger.instance().warn(f"Stv (={stv}) from pool does not have correct number of qubits (="
                                   f"{robot.num_of_qubits})!", show=False, from_pycui=False)
        return stv

    def produce_reward(self, rm: MyRandom) -> Optional[Collectible]:
        if self.__reward_factory is None:
            return None
        return self.__reward_factory.produce(rm)

    def copy_pool(self) -> List[StateVector]:
        return self.__pool.copy()



from abc import ABC
from typing import Optional, Tuple

from qrogue.game.logic.base import StateVector
from qrogue.game.logic.collectibles import Collectible, Instruction, instruction as gates, Score
from qrogue.util import PuzzleConfig
from .riddle import Riddle


class Boss(Riddle, ABC):
    """
    A special Enemy with specified target and reward.
    """

    __BOSS_ID: int = 0
    # we have to pass a seed to Riddle(), but we don't have instability so its value doesn't matter
    __PLACEHOLDER_SEED = 714985

    def __init__(self, id_: Optional[int], target: StateVector, input_: StateVector, reward: Collectible, edits: int):
        """
        Creates a boss enemy with a list of specified target and input StateVectors and a specified reward.
        :param id_: an integer unique per level to identify the target
        :param reward: the reward for winning against the boss
        """
        self.__index = 0

        if id_ is None:
            id_ = Boss.__BOSS_ID
            Boss.__BOSS_ID += 1
        super().__init__(id_, target, reward, Boss.__PLACEHOLDER_SEED, edits, input_, stable_probability=0)

    @property
    def flee_energy(self) -> int:
        return PuzzleConfig.BOSS_FLEE_ENERGY

    def flee_check(self) -> bool:
        return True


class EntanglementBoss(Boss):
    def __init__(self, reward: Collectible, edits: Optional[int] = None, num_of_qubits: Optional[int] = None,
                 entangled_qubits: Optional[Tuple[int, int]] = None):
        if edits is None: edits = 5
        if num_of_qubits is None: num_of_qubits = 2
        if entangled_qubits is None: entangled_qubits = (0, 1)
        input_stv = StateVector.create_zero_state_vector(num_of_qubits)
        target_stv = Instruction.compute_stv([
            gates.HGate().setup([entangled_qubits[0]]),
            gates.CXGate().setup([entangled_qubits[0], entangled_qubits[1]]),
        ], num_of_qubits)
        super().__init__(None, target_stv, input_stv, reward, edits)


class AntiEntangleBoss(Boss):
    def __init__(self, reward: Collectible, edits: Optional[int] = None):
        if edits is None: edits = 5

        comb_gate = gates.CombinedGate([
            gates.HGate().setup([0]), gates.CXGate().setup([0, 1]), gates.XGate().setup([0])
        ], 2, name="Anti Entanglement").setup([0, 1])
        target1 = Instruction.compute_stv([comb_gate], 2)
        target2 = Instruction.compute_stv([comb_gate, gates.XGate().setup([0])], 2)
        basis_states = StateVector.create_basis_states(num_of_qubits=2)
        puzzles = [
            (target1, basis_states[0]),
            (target1, basis_states[1]),
            (target2, basis_states[2]),
            (target2, basis_states[3]),
        ]
        super().__init__(None, puzzles[0][0], puzzles[0][1], reward, edits)


class ImaginaryBoss(Boss):
    def __init__(self, reward: Optional[Collectible], edits: Optional[int] = None):
        if reward is None: reward = Score(1000)
        if edits is None: edits = 14

        input_stv = StateVector.create_basis_states(num_of_qubits=2)[-1]
        gate_list = [gates.XGate().setup([1]), gates.HGate().setup([0]), gates.SGate().setup([1]),
                     gates.CXGate().setup([1, 0])]
        target_stv = Instruction.compute_stv(gate_list, num_of_qubits=2)

        super().__init__(None, target_stv, input_stv, reward, edits)



from typing import Optional, Dict, Any, Callable

from qrogue.util import MyRandom


class WaveFunction:
    def __init__(self, weights: Dict[Optional[Any], float], default_value: Any = None):
        self.__weights = weights
        self.__default_value = default_value
        self.__state: Optional[Any] = None

    @property
    def is_collapsed(self) -> bool:
        return self.__state is not None

    @property
    def state(self) -> Any:
        assert self.is_collapsed, "WaveFunction not yet collapsed!"
        return self.__state

    def adapt_weights(self, type_weights: Dict[Optional[Any], int]):
        """
        type_weights is only read, never written
        """
        if self.is_collapsed:
            return  # no need to adapt anything

        weight_sum = sum(type_weights.values())

        for key in self.__weights:
            if key in type_weights:
                norm_weight = type_weights[key] / weight_sum
                # for now the normalization will not be perfectly accurate
                # todo fix/come up with solution (floats?)
                # e.g. increase weight by 20% = multiply by 1.2
                self.__weights[key] = int(self.__weights[key] * (1 + norm_weight))
            else:
                self.__weights[key] = 0  # non-existing key is like a 0 weight

    def collapse(self, rand: MyRandom) -> Any:
        if len(self.__weights) == 0:
            self.__state = self.__default_value  # collapse to default_value if nothing else is possible
        if self.is_collapsed:
            return self.__state

        weight_sum = sum(self.__weights.values())
        rand_val = rand.get_int(0, weight_sum, "WaveFunction.collapse()")
        val = 0
        for key in self.__weights:
            weight = self.__weights[key]
            val += weight
            if rand_val < val:
                self.__state = key
                return self.__state
        from qrogue.util import Config
        Config.check_reachability("WaveFunction.collapse()@impossible-collapse")
        return None  # this should not be possible to happen

    def force_value(self, value: Any, is_assignable: Optional[Callable[[Any, Any], bool]] = None) -> bool:
        assert not self.is_collapsed, f"Forcing an already collapsed WaveFunction to {value}!"

        if is_assignable is None:
            def is_assignable(key_, value_) -> bool:
                return key_ == value_

        for key in self.__weights.keys():
            if is_assignable(key, value):
                self.__state = key
                return True
        return False

    def copy(self):
        wf = WaveFunction(self.__weights.copy(), self.__default_value)
        if self.is_collapsed:
            wf.force_value(self.__state)
        return wf

    def __str__(self):
        if self.is_collapsed:
            return f"WF ({self.__state})"
        else:
            text = f"WF (?"
            weight_sum = sum(self.__weights.values())
            for key in self.__weights:
                text += f"{str(key)} {(100 * self.__weights[key] / weight_sum):.0f}% | "
            text = text[:-2] + ")"
            return text
